from re import compile as re_compile
from smtplib import SMTP, SMTPServerDisconnected
from socket import gethostname
from typing import Optional

import dns.resolver as dns

DOMAIN_REGEX = re_compile(r'(?<=@)\[?([^\[\]]+)')


def _get_domain_from_email_address(email_address):
    try:
        return DOMAIN_REGEX.search(string=email_address)[1]
    except TypeError:
        raise ValueError('Invalid email address')
    except IndexError:
        raise ValueError('Invalid email address')


def _get_mx_records(domain: str) -> list:
    try:
        records = dns.query(domain, 'MX')
    except dns.NXDOMAIN:
        raise ValueError(f'Domain {domain} does not seem to exist')
    return [str(x.exchange) for x in records]


def mx_check(
    email_address: str, from_address: Optional[str] = None,
    helo_host: Optional[str] = None, smtp_timeout: int = 10
) -> Optional[bool]:
    """
    Return `True` if the host responds with a deliverable response code,
    `False` if not-deliverable.
    Also, return `None` if there was an error.
    """
    from_address = from_address or email_address
    host = helo_host or gethostname()

    smtp = SMTP(timeout=smtp_timeout)
    smtp.set_debuglevel(0)

    domain = _get_domain_from_email_address(email_address)
    try:
        mx_records = _get_mx_records(domain)
    except ValueError:
        return False

    for mx_record in mx_records:
        smtp.connect(mx_record)
        smtp.helo(host)
        smtp.mail(from_address)
        code, message = smtp.rcpt(email_address)
        try:
            smtp.quit()
        except SMTPServerDisconnected:
            return None
        if code == 250:
            return True
        if 400 <= code <= 499:
            # Ambigious return code, can be graylist, or temporary
            # problems
            return None
    return False



from ipaddress import IPv4Address, IPv6Address
from os.path import dirname, join
from re import IGNORECASE
from re import compile as re_compile
from typing import Optional

SetOrNone = Optional[set]


def _validate_ipv4_address(value: str):
    try:
        IPv4Address(address=value)
    except ValueError:
        return False
    return True


def _validate_ipv6_address(value: str) -> bool:
    """
    Return whether or not the `ip_str` string is a valid IPv6 address.
    """
    try:
        IPv6Address(address=value)
    except ValueError:
        return False
    return True


def _validate_ipv46_address(value: str) -> bool:
    if _validate_ipv4_address(value):
        return True
    return _validate_ipv6_address(value)


class EmailValidator(object):
    'Slightly adjusted email regex checker from the Django project.'
    user_regex = re_compile(
        # dot-atom
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*\Z"
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013'
        r'\014\016-\177])*"\Z)', IGNORECASE)
    domain_regex = re_compile(
        # max length for domain name labels is 63 characters per RFC 1034
        r'((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+)'
        r'(?:[A-Z0-9-]{2,63}(?<!-))\Z', IGNORECASE)
    literal_regex = re_compile(
        # literal form, ipv4 or ipv6 address (SMTP 4.1.3)
        r'\[([A-f0-9:\.]+)\]\Z', IGNORECASE)
    domain_whitelist = frozenset('localhost')
    domain_blacklist = frozenset()

    def __init__(
            self, whitelist: SetOrNone = None, blacklist: SetOrNone = None):
        self.domain_whitelist = set(whitelist) \
            if whitelist else self.domain_whitelist
        self._load_blacklist(blacklist=blacklist)

    def _load_blacklist(self, blacklist: SetOrNone = None):
        'Load our blacklist.'
        self.domain_blacklist = set(blacklist) \
            if blacklist else self.domain_blacklist
        path = join(dirname(__file__), 'lib', 'blacklist.txt')
        try:
            with open(path) as fd:
                lines = fd.readlines()
        except FileNotFoundError:
            return
        self.domain_blacklist = self.domain_blacklist.union(
            x.strip() for x in lines)

    def __call__(self, value: str, use_blacklist: bool = True) -> bool:
        if not value or '@' not in value:
            return False

        user_part, domain_part = value.rsplit('@', 1)

        if not self.user_regex.match(user_part):
            return False

        if domain_part in self.domain_whitelist:
            return True
        if domain_part in self.domain_blacklist:
            return False

        if not self.validate_domain_part(domain_part):
            # Try for possible IDN domain-part
            try:
                domain_part = domain_part.encode('idna').decode('ascii')
            except UnicodeError:
                pass
            else:
                if self.validate_domain_part(domain_part):
                    return True
            return False
        return True

    def validate_domain_part(self, domain_part):
        if self.domain_regex.match(domain_part):
            return True

        literal_match = self.literal_regex.match(domain_part)
        if literal_match:
            ip_address = literal_match.group(1)
            return _validate_ipv46_address(ip_address)
        return False


regex_check = EmailValidator()



from typing import Optional

from .mx_check import mx_check
from .regex_check import regex_check


def validate_email(
        email_address: str, check_regex: bool = True, check_mx: bool = True,
        from_address: Optional[str] = None, helo_host: Optional[str] = None,
        smtp_timeout: int = 10, use_blacklist: bool = True) -> Optional[bool]:
    """
    Return `True` or `False` depending if the email address exists
    or/and can be delivered.

    Return `None` if the result is ambigious.
    """
    if check_regex and not regex_check(
            value=email_address, use_blacklist=use_blacklist):
        return False
    if not check_mx:
        return True
    return mx_check(
        email_address=email_address, from_address=from_address,
        helo_host=helo_host, smtp_timeout=smtp_timeout)



from os.path import join
from urllib.request import urlopen

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

blacklist_url = (
    'https://raw.githubusercontent.com/martenson/disposable-email-domains/'
    'master/disposable_email_blocklist.conf')


class PostBuildPyCommand(build_py):
    'Post-installation for build_py'

    def run(self):
        if self.dry_run:
            return super().run()
        with urlopen(url=blacklist_url) as fd:
            content = fd.read().decode('utf-8')
        target_dir = join(self.build_lib, 'validate_email/lib')
        self.mkpath(target_dir)
        with open(join(target_dir, 'blacklist.txt'), 'w') as fd:
            fd.write(content)
        super().run()


setup(
    name='py3-validate-email',
    version='0.1.3',
    packages=find_packages(exclude=['tests']),
    install_requires=['dnspython'],
    author='László Károlyi',
    author_email='laszlo@karolyi.hu',
    description='Email validator with regex and SMTP checking.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    keywords='email validation verification mx verify',
    url='http://github.com/karolyi/py3-validate-email',
    cmdclass=dict(build_py=PostBuildPyCommand),
    license='LGPL',
)
